﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Globalization;

namespace GPSLib
{
    /*
     * $GPGGA,123519,4807.038,N,01131.000,E,1,08,0.9,545.4,M,46.9,M,,*47
     0  GGA          Global Positioning System Fix Data
     1   123519       Fix taken at 12:35:19 UTC
     2-3 4807.038,N   Latitude 48 deg 07.038' N
     4-5 01131.000,E  Longitude 11 deg 31.000' E
     6 1            Fix quality: 0 = invalid
                               1 = GPS fix (SPS)
                               2 = DGPS fix
                               3 = PPS fix
                               4 = Real Time Kinematic
                               5 = Float RTK
                               6 = estimated (dead reckoning) (2.3 feature)
                               7 = Manual input mode
                               8 = Simulation mode
     7 08           Number of satellites being tracked
     8 0.9          Horizontal dilution of position
     9-10 545.4,M      Altitude, Meters, above mean sea level
     11-12 46.9,M       Height of geoid (mean sea level) above WGS84
                      ellipsoid
     (empty field) time in seconds since last DGPS update
     (empty field) DGPS station ID number
     *47          the checksum data, always begins with *
    */
    public class NMEADataGGA
    {
        private static CultureInfo CultureUS = CultureInfo.CreateSpecificCulture("en-US");
        private string[] source;
        public double altitude;
        public int fixQuality;

        public bool isValid()
        {
            return (fixQuality > 0);
        }

        public NMEADataGGA(string[] data)
        {
            source = data;
            altitude = double.Parse(data[9], CultureUS);
            fixQuality = Int16.Parse(data[6]);
        }
    }

    /*
     $GPGSA,A,3,04,05,,09,12,,,24,,,,,2.5,1.3,2.1*39

    Where:
     GSA      Satellite status
     0 A        Auto selection of 2D or 3D fix (M = manual) 
     1 3        3D fix - values include: 1 = no fix
                                       2 = 2D fix
                                       3 = 3D fix
     2-13 04,05... PRNs of satellites used for fix (space for 12) 
     14 2.5      PDOP (dilution of precision) 
     15 1.3      Horizontal dilution of precision (HDOP) 
     16 2.1      Vertical dilution of precision (VDOP)
     *39      the checksum data, always begins with *
    */
    public class NMEADataGSA
    {
        private string[] source;
        public NMEADataGSA(string[] data)
        {
            source = data;
        }
    }

    /*
      $GPGSV,2,1,08,01,40,083,46,02,17,308,41,12,07,344,39,14,22,228,45*75

    Where:
         0 GSV          Satellites in view
         1 2            Number of sentences for full data
         2 1            sentence 1 of 2
         3 08           Number of satellites in view

         4 01           Satellite PRN number
         5 40           Elevation, degrees
         6 083          Azimuth, degrees
         7 46           SNR - higher is better
                           for up to 4 satellites per sentence
         8 *75          the checksum data, always begins with *
    */
    public class NMEADataGSV
    {
        private string[] source;
        public NMEADataGSV(string[] data)
        {
            source = data;
        }
    }

    /*
     $GPRMC,123519,A,4807.038,N,01131.000,E,022.4,084.4,230394,003.1,W*6A

     Where:
        0 RMC          Recommended Minimum sentence C
        1 123519       Fix taken at 12:35:19 UTC
        2 A            Status A=active or V=Void.
        3-4 4807.038,N   Latitude 48 deg 07.038' N
        5-6 01131.000,E  Longitude 11 deg 31.000' E
        7 022.4        Speed over the ground in knots
        8 084.4        Track angle in degrees True
        9 230394       Date - 23rd of March 1994
        10-11 003.1,W      Magnetic Variation
         *6A          The checksum data, always begins with *
    */
    public class NMEADataRMC
    {
        private static CultureInfo CultureUS = CultureInfo.CreateSpecificCulture("en-US");
        private string[] source;

        public enum StatusValues { Active, Void };
        public StatusValues status { get; private set; }
        public DateTime time { get; private set; }
        public String magneticVariation { get; private set; }
        public String latitude {  get; private set; }
        public String longitude {  get; private set; }
        public double heading {  get; private set; }
        public double groundSpeedKnots {  get; private set; }

        public double groundSpeedKMH()
        {
            return groundSpeedKnots * 1.852;
        }

        public NMEADataRMC(string[] data)
        {
            source = data;

            // Status
            switch (data[2])
            {
                case "A":
                    status = StatusValues.Active;
                    break;
                case "V":
                default:
                    status = StatusValues.Void;
                    break;
            }
            // Time
            string dateString = data[9].Substring(4, 2) + "-" + data[9].Substring(2, 2) + "-" + data[9].Substring(0, 2);
            string timeString = data[1].Substring(0, 2) + ":" + data[1].Substring(2, 2) + ":" + data[1].Substring(4, 2);
            string dateTimeString = "20"+dateString + " " + timeString + "Z";
            time = DateTime.ParseExact(dateTimeString,"u",System.Globalization.CultureInfo.InvariantCulture);
         
            // Magnetic
            magneticVariation = data[10] + data[11];
            // Lat
            latitude = data[3] + data[4];
            // Long
            longitude = data[5] + data[6];
            // Track
            heading = double.Parse(data[8], CultureUS);
            // Speed
            groundSpeedKnots = double.Parse(data[7], CultureUS);
        }
    }


    public class NMEAData
    {
        public enum types { NONE = 0, 
                            GGA = 1,
                            GSA = 2, 
                            GSV = 4,
                            RMC = 8, 
                        };
        public types type { get; private set; }
        private int checksum = 0;
        private string dataString = null;
        private string[] dataArr;

        public NMEADataGGA gga { get; private set; }
        public NMEADataGSA gsa { get; private set; }
        public NMEADataGSV gsv { get; private set; }
        public NMEADataRMC rmc { get; private set; }

        public NMEAData()
        {
            type = types.NONE;
        }

        public NMEAData(string line)
        {
            type = types.NONE;
            line.Trim();
            String pat = @"^\$(.*)\*([0-9A-F]{2})$";
            Regex r = new Regex(pat);
            Match m = r.Match(line);
            if (m.Success)
            {
                dataString = m.Groups[1].ToString();
                string checkSumString = m.Groups[2].ToString();
                checksum = int.Parse(checkSumString, System.Globalization.NumberStyles.HexNumber);
                if (checksum == calcChecksum(dataString))
                {
                    // We split on commas, but actually there also is one last
                    // field between last comma and the asterisk. This needs to
                    // be fixed if we want to use that value
                    dataArr = dataString.Split(',');
                    
                    // Fill in the blanks
                    for (int i = 0; i < dataArr.Length; i++)
                    {
                        if (dataArr[i] == "")
                        {
                            dataArr[i] = "0"; 
                        }
                    }

                    switch (dataArr[0])
                    {
                        case "GPGGA":
                            type = types.GGA;
                            break;
                        case "GPGSA":
                            type = types.GSA;
                            break;
                        case "GPGSV":
                            type = types.GSV;
                            break;
                        case "GPRMC":
                            type = types.RMC;
                            break;
                        default:
                            type = types.NONE;
                            break;
                    }

                    switch (type)
                    {
                        case types.GGA:
                            gga = new NMEADataGGA(dataArr);
                            break;
                        case types.GSA:
                            gsa = new NMEADataGSA(dataArr);
                            break;
                        case types.GSV:
                            gsv = new NMEADataGSV(dataArr);
                            break;
                        case types.RMC:
                            rmc = new NMEADataRMC(dataArr);
                            break;
                        default:
                            break;
                    }
                }
            }
        }


        private static int calcChecksum(String str)
        {
            int res = 0;
            foreach (char c in str.ToCharArray())
            {
                res ^= c;
            }
            return res;
        }

    }
}
